package c10quality.lecture.person;

import java.util.Calendar;
import java.util.GregorianCalendar;
import java.util.Locale;

/**
 * This class models a person using name, gender and birthdate. It is extended
 * with exceptions, unit testing and enumerations.
 * 
 * Person class invariant: (myFirstName != null) && (!myFirstName.equals("")) &&
 * (myLastName != null) && (myLastName.equals("")) && (myGender == Gender.MALE
 * || myGender == Gender.FEMALE) && myBirthdate != null
 * 
 * @author kvlinden
 * @version Spring, 2012 (based on the Chapter 9 version)
 */
public class Person {
	
	public static final int VOTING_AGE = 18;

	public enum Gender {
		MALE, FEMALE
	};

	private String myFirstName;
	private String myLastName;
	private Gender myGender;
	private GregorianCalendar myBirthdate;

	/**
	 * This constructor constructs a new default person named John Doe.
	 */
	public Person() {
		myFirstName = "John";
		myLastName = "Doe";
		myGender = Gender.MALE;
		myBirthdate = new GregorianCalendar();
	}

	/**
	 * This constructor constructs a new person characterized by the parameter
	 * values.
	 * 
	 * @param firstName
	 *            the first name of this person object (cannot be empty or null)
	 * @param lastName
	 *            the surname of this person object
	 * @param gender
	 *            the gender of this person object (M or F)
	 * @param birthdate
	 *            the date of birth of this person object
	 * @throws Exception
	 */
	public Person(String firstName, String lastName, Gender gender, GregorianCalendar birthdate)
			throws Exception {
		if (isValidInformation(firstName, lastName, gender, birthdate)) {
			myFirstName = firstName;
			myLastName = lastName;
			myGender = gender;
			myBirthdate = birthdate;
		} else {
			throw new Exception("Invalid personal information: " + firstName + " " + lastName + " "
					+ gender + " " + birthdate);
		}
	}

	/**
	 * @return the first name of this person object
	 */
	public String getFirstName() {
		return myFirstName;
	}

	/**
	 * @return the surname name of this person object
	 */
	public String getLastName() {
		return myLastName;
	}

	/**
	 * @return the gender of this person object (M or F)
	 */
	public Gender getGender() {
		return myGender;
	}

	/**
	 * @return the birthdate of this person object
	 */
	public GregorianCalendar getBirthdate() {
		return myBirthdate;
	}
	
	/**
	 * @return true if this person is of voting age
	 */
	public boolean canVote() {
		return getAge() >= VOTING_AGE;
	}

	/**
	 * @return the age in years of this person object, -1 if no birthdate is
	 *         known
	 */
	public int getAge() {
		if (myBirthdate == null) {
			return -1;
		} else {
			// Compute "raw" age in years.
			Calendar now = new GregorianCalendar();
			int age = now.get(Calendar.YEAR) - myBirthdate.get(Calendar.YEAR);

			// Subtract one year if my birthday hasn't come yet this year.
			if ((myBirthdate.get(Calendar.MONTH) > now.get(Calendar.MONTH))
					|| (myBirthdate.get(Calendar.MONTH) == now.get(Calendar.MONTH) && myBirthdate
							.get(Calendar.DAY_OF_MONTH) > now.get(Calendar.DAY_OF_MONTH))) {
				age--;
			}

			return age;
		}
	}

	/**
	 * Change the name of this person object if the invariant is still
	 * satisfied.
	 * 
	 * @param firstName
	 *            the proposed new first name
	 * @param lastName
	 *            the proposed new surname
	 * @throws Exception
	 */
	public void setName(String firstName, String lastName) throws Exception {
		if (isValidInformation(firstName, lastName, myGender, myBirthdate)) {
			myFirstName = firstName;
			myLastName = lastName;
		} else {
			throw new Exception("Invalid personal information: " + firstName + " " + lastName + " "
					+ myGender + " " + myBirthdate);
		}
	}

	/**
	 * Determine if I am older than the given person.
	 * 
	 * @param otherPerson
	 *            the given person to compare against
	 * @return true if I am older, false otherwise
	 */
	public boolean isOlder(Person otherPerson) {
		return myBirthdate.before(otherPerson.getBirthdate());
	}

	/**
	 * Ensure that the class invariant (see the class documentation above) is
	 * satisfied.
	 * 
	 * @param firstName
	 *            the proposed first name
	 * @param lastName
	 *            the proposed surname
	 * @param gender
	 *            the proposed gender
	 * @param birthdate
	 *            the proposed birthdate
	 * @return true if the invariant is satisfied, false otherwise
	 */
	private static boolean isValidInformation(String firstName, String lastName, Gender gender,
			GregorianCalendar birthdate) {
		// Make sure that names aren't null.
		if ((firstName == null) || (lastName == null)) {
			return false;
		}
		// Make sure that names aren't empty strings.
		if ((firstName.equals("")) || (lastName.equals(""))) {
			return false;
		}
		// Make sure that the gender is either male or female (accept either
		// case).
		if ((gender != Gender.FEMALE) && (gender != Gender.MALE)) {
			return false;
		}
		// Make sure birthdate is not null.
		if (birthdate == null) {
			return false;
		}
		return true;
	}

	/**
	 * Construct a string naming the person, which includes a Mr or Ms honorific
	 * as appropriate.
	 */
	public String toString() {
		String honorific;
		if (myGender == Gender.MALE) {
			honorific = "Mr.";
		} else {
			honorific = "Ms.";
		}
		return honorific + " " + myFirstName + " " + myLastName;
	}

}
